/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DielectricsRemoteMatlabReq.h"
#include "Spectrums.h"
#include <InstrumentManager/InstrumentManager.h>
#include <boost/lexical_cast.hpp>
#include <fstream>
#include <iostream>
#include <common/base/ServerProperties.h>
#include "controllers/common/family/Families.h"
#include "FrequencyRequestMessages.pb.h"

namespace dielectrics {

using namespace std;
using namespace common;
using namespace boost;
using namespace cameo;

const std::string RemoteMatlabReq::TYPE = "dielectrics_remote_matlab";
const std::string RemoteMatlabReq::MATLAB_APPLICATION = "dielectrics-matlab";
const std::string RemoteMatlabReq::MATLAB_STOP_APPLICATION = "dielectrics-matlab-stop";
const std::string RemoteMatlabReq::RESPONDER_NAME = "dielectrics_controller";
const std::string RemoteMatlabReq::PUBLISHER_NAME = "dielectrics_pub";

const std::string RemoteMatlabReq::RUN_SPECTRUMS_TEST = "Run Spectrums Test";
const std::string RemoteMatlabReq::RUN_SPECTRUMS = "Run Spectrums";
const std::string RemoteMatlabReq::RUN_SPECTRUMS_TIME = "Run Spectrums Time";
const std::string RemoteMatlabReq::RUN_SINGLE_FREQUENCY = "Run Single Frequency";

RemoteMatlabReq::RemoteMatlabReq(const std::string& name) :
	ExperimentController(name), controller::Start(this), controller::Raz(this), Test(this) {

	setFamily(family::ACQUISITION, family::POLARIZATIONS);

	serverEndpoint.init(this, SAVE, "cameo_server");
	initialized.init(this, NOSAVE, "initialized");
	applicationState.init(this, NOSAVE, "matlab_state");

	values.init(this, NOSAVE, "values");

	requestTypeValues.init(this, NOSAVE, "request_type_values");

	requestTypeValues.resize(4);
	requestTypeValues.set(0, RUN_SPECTRUMS_TEST);
	requestTypeValues.set(1, RUN_SPECTRUMS);
	requestTypeValues.set(2, RUN_SPECTRUMS_TIME);
	requestTypeValues.set(3, RUN_SINGLE_FREQUENCY);

	requestType.init(this, NOSAVE, "request_type", "Request");
	requestType.setEnumeratedLabels(requestTypeValues);
	requestType.setEnumeratedValues(requestTypeValues);

	requestType = requestTypeValues.get(0);

	fr.init(this, NOSAVE, "fr");
	frMin.init(this, NOSAVE, "frMin");
	frMax.init(this, NOSAVE, "frMax");
	nMax.init(this, NOSAVE, "nMax");
	time.init(this, NOSAVE, "time");
	comment.init(this, NOSAVE, "comment");

	actualN.init(this, NOSAVE, "N");
	actualTime.init(this, NOSAVE, "actual_time");
	actualRun.init(this, NOSAVE, "actual_run");

	fr = 0;
	frMin = 1E1;
	frMax = 1E6;
	nMax = 5;
	time = 0;
	comment = "a comment";
}

RemoteMatlabReq::RemoteMatlabReq(const RemoteMatlabReq& controller) :
	ExperimentController(controller), controller::Start(this), controller::Raz(this), Test(this) {
}

RemoteMatlabReq::~RemoteMatlabReq() {
	raz();
}

void RemoteMatlabReq::postConfiguration() {

	if (m_server.get() == 0) {
		// Create a new server if it is not already created.
		if (serverEndpoint() == "") {
			m_server.reset(new Server(application::This::getServer().getEndpoint()));
		}
		else {
			m_server.reset(new Server(serverEndpoint()));
		}
	}

	// Do not initialise if the controller is disabled.
	if (isEnabled()) {
		initialized = initApplication();
	}
}

void RemoteMatlabReq::setEnabled(bool all, bool postconf) {

	ExperimentController::setEnabled(all);

	// Enable the spectrums controller.
	for (vector<Spectrums *>::const_iterator c = m_spectrumsControllers.begin(); c != m_spectrumsControllers.end(); ++c) {
		(*c)->setEnabled(false, false);
	}
}

void RemoteMatlabReq::setDisabled(bool all) {

	raz();
	ExperimentController::setDisabled(all);

	// Disable the spectrums controller.
	for (vector<Spectrums *>::const_iterator c = m_spectrumsControllers.begin(); c != m_spectrumsControllers.end(); ++c) {
		(*c)->setDisabled(false);
	}
}

void RemoteMatlabReq::addSpectrumsController(Spectrums * controller) {
	m_spectrumsControllers.push_back(controller);
}

bool RemoteMatlabReq::initApplication() {

	// Do not initialize if it is already done
	if (initialized()) {
		return true;
	}

	try {
		// Start the Matlab server
		if (!m_server->isAvailable(10000)) {
			cout << "Matlab server is not available" << endl;
			return false;
		}

		cout << "Matlab server is connected to " << getName() << endl;

		m_matlabApplication = m_server->connect(MATLAB_APPLICATION);

		cout << "connected application" << endl;

		if (m_matlabApplication->exists()) {
			// The application exists from a previous server session
			m_matlabApplication->stop();
			application::State state = m_matlabApplication->waitFor();
			cout << "Terminated matlab application " << state << endl;
		}

		m_matlabApplication = m_server->start(MATLAB_APPLICATION);

		if (!m_matlabApplication->exists()) {
			cout << "No matlab application" << endl;
			return false;
		}

		applicationState = m_matlabApplication->waitFor(application::STARTING);

		// First delete the requester if it exists.
		m_requester.reset(0);

		// Create a new requester.
		m_requester = application::Requester::create(*m_matlabApplication, RESPONDER_NAME);

		if (m_requester.get() == 0) {
			cout << "requester error" << endl;
			return false;
		}

		// Create a new subscriber.
		m_subscriber = application::Subscriber::create(*m_matlabApplication, PUBLISHER_NAME);

		// Start the subscriber loop
		m_subscriberThread.reset(new thread(bind(&RemoteMatlabReq::subscriberLoop, this)));

		// Application initialized.
		initialized = true;

		applicationState = m_matlabApplication->waitFor(application::RUNNING);

		cout << "Matlab application initialized" << endl;

		return true;

	} catch (const std::exception & e) {
		// Currently an exception can occur during isAvailable.
		cout << "Unexpected exception during matlab connection: " << e.what() << endl;
	}

	return false;
}

void RemoteMatlabReq::start() {

	if (!initApplication()) {
		return;
	}

	commandStatus.setRunning();

	// Serialize the request.
	proto::RunSpectrumsRequest request;

	request.set_fr(fr());
	request.set_frmin(frMin());
	request.set_frmax(frMax());
	request.set_nmax(nMax());
	request.set_time(time());
	request.set_comment(comment());

	if (requestType() == RUN_SPECTRUMS) {
		request.set_type(proto::RunSpectrumsRequest_Type_RUN);
	}
	else if (requestType() == RUN_SPECTRUMS_TIME) {
		request.set_type(proto::RunSpectrumsRequest_Type_RUN_TIME);
	}
	else if (requestType() == RUN_SINGLE_FREQUENCY) {
		request.set_type(proto::RunSpectrumsRequest_Type_RUN_SINGLE);
	}
	else if (requestType() == RUN_SPECTRUMS_TEST) {
		request.set_type(proto::RunSpectrumsRequest_Type_TEST);
	}

	// Send the message.
	m_requester->sendBinary(request.SerializeAsString());

	cout << getName() << " sent run request " << requestType() << endl;

	// Wait for the response synchronously.
	// Note that responses can be processed asynchronously if the server is able to.
	string data;
	m_requester->receiveBinary(data);

	proto::FrequencyResponse response;
	response.ParseFromString(data);

	cout << "received response " << response.responseid() << endl;

	delete [] values.get();
	int32* newArray = new int32[response.values_size()];

	for (int i = 0; i < response.values_size(); ++i) {
		newArray[i] = response.values(i);
	}

	values.set(newArray);
	values.setSize(response.values_size());

	sleep(1);

	commandStatus.setIdle();
}

void RemoteMatlabReq::raz() {

	// Reset if the controller was initialized.
	if (initialized()) {

		// Stop the subscriber
		m_subscriber->cancel();

		// The subscriber thread terminates
		m_subscriberThread->join();

		// First stop the application.
		stopApplication();

		// Restart the application.
		initialized = false;
	}
}

void RemoteMatlabReq::test() {

	for (int i = 0; i < 100; ++i) {
		raz();
		initApplication();
	}
}

void RemoteMatlabReq::stopRun() {

	auto_ptr<cameo::application::Instance> matlabStopApplication = m_server->start(MATLAB_STOP_APPLICATION);
	application::State state = matlabStopApplication->waitFor();
}

void RemoteMatlabReq::stopApplication() {

	if (!initialized()) {
		return;
	}

	// Exit immediately if the server is not available.
	if (!m_server->isAvailable(10000)) {
		return;
	}

	// Serialize the request.
	proto::RunSpectrumsRequest request;

	request.set_fr(0);
	request.set_frmin(0);
	request.set_frmax(0);
	request.set_nmax(0);
	request.set_time(0);
	request.set_comment("");

	request.set_type(proto::RunSpectrumsRequest_Type_STOP);

	// Send the message.
	m_requester->sendBinary(request.SerializeAsString());

	cout << getName() << " sent stop request" << endl;

	// Wait for the response synchronously.
	// Note that responses can be processed asynchronously if the server is able to.
	string data;
	m_requester->receiveBinary(data);

	proto::FrequencyResponse response;
	response.ParseFromString(data);

	cout << "received response " << response.responseid() << endl;

	// Stop the remote application.
	m_matlabApplication->stop();

	// Wait for the termination
	applicationState = m_matlabApplication->waitFor();

	cout << "Matlab application terminated with state " << application::toString(applicationState()) << endl;
}

void RemoteMatlabReq::subscriberLoop() {

	// Loop on events
	string data;
	while (m_subscriber->receiveBinary(data)) {

		// Parse the status.
		proto::RunSpectrumsStatus status;
		status.ParseFromString(data);

		// No need to read the type.
		actualN = status.n();
		actualTime = status.time();
		actualRun = status.run();
	}
}

}
