/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "RunSpectrums.h"

namespace dielectrics {

using namespace std;
using namespace common;
using namespace cameo;

const std::string RunSpectrums::TYPE = "dielectrics_run_spectrums";

RunSpectrums::RunSpectrums(const std::string& name) :
	Spectrums(name), controller::Stoppable(this), controller::Raz(this) {

	fr.init(this, SAVE, "fr", "Freq");
	frMin.init(this, SAVE | SPY, "frMin", "Freq Min");
	frMax.init(this, SAVE | SPY, "frMax", "Freq Max");
	N.init(this, SAVE | SPY, "actual_N", "wanted_N", "N");
	time.init(this, SAVE, "actual_time", "wanted_time", "Time");
	comment.init(this, SAVE, "comment", "Comment");
	actualRun.init(this, NOSAVE | SPY, "run", "Run");
}

RunSpectrums::~RunSpectrums() {
}

void RunSpectrums::postConfiguration() {

	Spectrums::postConfiguration();

	registerUpdater(matlabController->applicationState, &RunSpectrums::updateApplicationState, this);
	registerStatus(matlabController, &RunSpectrums::updateState, this);

	registerUpdater(matlabController->actualN, &RunSpectrums::updateN, this);
	registerUpdater(matlabController->actualTime, &RunSpectrums::updateTime, this);
	registerUpdater(matlabController->actualRun, &RunSpectrums::updateRun, this);

	updateApplicationState();
}

void RunSpectrums::updateApplicationState() {

	int32 state = matlabController->applicationState();

	if (state == application::STARTING) {
		commandStatus.setWarning();
	}
	else if (state == application::RUNNING) {
		commandStatus.setIdle();
	}
	else {
		commandStatus.setError();
	}
}

void RunSpectrums::updateState() {

	if (isStarted() && matlabController->commandStatus.isRunning()) {
		commandStatus.setRunning();
	}
	else {
		commandStatus.setIdle();
	}
}

void RunSpectrums::updateN() {

	if (isStarted()) {
		N = matlabController->actualN();
	}
}

void RunSpectrums::updateTime() {

	if (isStarted()) {
		time = matlabController->actualTime();
	}
}

void RunSpectrums::updateRun() {

	if (isStarted()) {
		actualRun = matlabController->actualRun();
	}
}

void RunSpectrums::start() {

	matlabController->requestType = RemoteMatlabReq::RUN_SPECTRUMS;
	matlabController->frMin = frMin();
	matlabController->frMax = frMax();
	matlabController->nMax = N.setpoint();
	matlabController->comment = comment();

	matlabController->startCommand();

	LogStream logStream = log(Level::s_Info);
	logStream << "Freq min " << frMin << " Freq Max " << frMax << " N Max " << N.setpoint << " Comment '" << comment << "' ";

	logRunNumbers(logStream);
	logStream << endlog;
}

void RunSpectrums::stop() {
}

void RunSpectrums::raz() {

	matlabController->razCommand();

	log(Level::s_Info) << "Stopped remote application" << endlog;
}

void RunSpectrums::logRunNumbers(LogStream& logStream) {

	logStream << "run numbers [";

	int32 size = matlabController->values.getSize();
	int32* values = matlabController->values.get();
	for (int32 i = 0; i < size - 1; ++i) {
		logStream << values[i] << ", ";
	}

	if (size > 0) {
		logStream << values[size - 1] << "]";
	}
}

}
