/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <cmath>

#include "CAENSYx527Controller.h"
#include "controllers/common/family/Families.h"
#include "CAENSYx527Def.h"

using namespace std;

namespace caen_syx527 {

const string CAENSYx527Controller::TYPE = "caen_syx527_controller";

const string CAENSYx527Controller::CURRENT_UNIT = "uA";
const string CAENSYx527Controller::VOLTAGE_UNIT = "V";

/*
 * Constructor
 */
CAENSYx527Controller::CAENSYx527Controller(const string& name) :
	ExperimentController(name), controller::Stoppable(this), controller::Read(this), On(this), Off(this),Select(this),Set(this) {

	setFamily(family::VOLTAGE);

	// Init property
	ntotchannel.init(this, NOSAVE, "ntotchannel");
	islot.init(this, NOSAVE, "islot");
	ichannel.init(this, NOSAVE, "ichannel");
	namechannel.init(this, NOSAVE, "namechannel");
	i0set.init(this, NOSAVE | SPY, "i0set");
	v0set.init(this, NOSAVE | SPY, "v0set");
	imon.init(this, NOSAVE | SPY, "imon");
	vmon.init(this, NOSAVE | SPY, "vmon");
	power.init(this, NOSAVE, "power");
	statuschannel.init(this, NOSAVE, "statuschannel");
	rampup.init(this, NOSAVE, "rampup");
	rampdown.init(this, NOSAVE, "rampdown");
	isselected.init(this, NOSAVE, "isselected");

	selectchoice.init(this,NOSAVE,"selectchoice");
	selectslot.init(this,NOSAVE,"selectslot");
	selectstring.init(this,NOSAVE,"selectstring");
	setparam.init(this,NOSAVE,"setparam");
	setvalue.init(this,NOSAVE,"setvalue");


	vmax.init(this,SAVE,"vmax");
	deltav.init(this,NOSAVE,"deltav");

	statusall.init(this,NOSAVE,"statusall");
	statusconnection.init(this,NOSAVE,"statusconnection");

	m_Driver.init(this, "driver");

	m_StartActivated = false;

	registerFunction(TYPE);


}

CAENSYx527Controller::CAENSYx527Controller(const CAENSYx527Controller& controller) :
	ExperimentController(controller), controller::Stoppable(this), controller::Read(this), On(this), Off(this),Select(this),Set(this) {

	ntotchannel.copy(this, controller.ntotchannel);
	islot.copy(this, controller.islot);
	ichannel.copy(this, controller.ichannel);
	namechannel.copy(this, controller.namechannel);
	i0set.copy(this, controller.i0set);
	v0set.copy(this, controller.v0set);
	imon.copy(this, controller.imon);
	vmon.copy(this, controller.vmon);
	power.copy(this, controller.power);
	statuschannel.copy(this, controller.statuschannel);
	rampup.copy(this, controller.rampup);
	rampdown.copy(this, controller.rampdown);
	isselected.copy(this, controller.isselected);

	selectchoice.copy(this, controller.selectchoice);
	selectslot.copy(this, controller.selectslot);
	selectstring.copy(this, controller.selectstring);
	setparam.copy(this, controller.setparam);
	setvalue.copy(this, controller.setvalue);
	vmax.copy(this, controller.vmax);
	deltav.copy(this, controller.deltav);
	statusall.copy(this, controller.statusall);
	statusconnection.copy(this, controller.statusconnection);

	m_Driver.copy(controller.m_Driver);

	m_StartActivated = false;
}

/*
 * Destructor
 */
CAENSYx527Controller::~CAENSYx527Controller() {
	// Empty
}

void CAENSYx527Controller::postConfiguration() {
	registerRefresher(ntotchannel,&CAENSYx527Controller::refreshNTotalChannelProperty,this);
	ntotchannel=m_Driver->ntotchannel();
	for(int i=0;i<ntotchannel();i++){
			islot.set(i,m_Driver->islot(i));
			ichannel.set(i,m_Driver->ichannel(i));
			namechannel.set(i,m_Driver->namechannel(i));
			v0set.set(i,m_Driver->v0set(i));
			i0set.set(i,m_Driver->i0set(i));
			vmon.set(i,m_Driver->vmon(i));
			imon.set(i,m_Driver->imon(i));
			rampup.set(i,m_Driver->rampup(i));
			rampdown.set(i,m_Driver->rampdown(i));
			statuschannel.set(i,m_Driver->statuschannel(i));
			power.set(i,m_Driver->power(i));
			isselected.set(i,false);
			//cout << "Controller " << islot(i) << " " << namechannel(i) << " " << v0set(i) << " " << isselected(i) << endl;
	}
    deltav=0;
    statusall=true;
    statusconnection=true;

	registerRefresher(namechannel, &CAENSYx527Controller::refreshChannelNameProperty, this);
	registerRefresher(power, &CAENSYx527Controller::refreshPowerProperty, this);
	registerRefresher(v0set, &CAENSYx527Controller::refreshV0SetProperty, this);
	registerRefresher(i0set, &CAENSYx527Controller::refreshI0SetProperty, this);
	registerRefresher(rampup, &CAENSYx527Controller::refreshRampUpProperty, this);
	registerRefresher(rampdown, &CAENSYx527Controller::refreshRampDownProperty, this);
	registerRefresher(vmax, &CAENSYx527Controller::refreshVMaxProperty, this);
	registerRefresher(deltav, &CAENSYx527Controller::refreshDeltaVProperty, this);

	registerUpdater(m_Driver->namechannel, &CAENSYx527Controller::updateChannelName, this);
	registerUpdater(m_Driver->v0set, &CAENSYx527Controller::updateV0Set, this);
	registerUpdater(m_Driver->i0set, &CAENSYx527Controller::updateI0Set, this);
	registerUpdater(m_Driver->vmon, &CAENSYx527Controller::updateVMon, this);
	registerUpdater(m_Driver->imon, &CAENSYx527Controller::updateIMon, this);
	registerUpdater(m_Driver->power, &CAENSYx527Controller::updatePower, this);
	registerUpdater(m_Driver->rampup, &CAENSYx527Controller::updateRampUp, this);
	registerUpdater(m_Driver->rampdown, &CAENSYx527Controller::updateRampDown, this);
	registerUpdater(m_Driver->statuschannel, &CAENSYx527Controller::updateChannelStatus, this);
	registerUpdater(m_Driver->statusconnection, &CAENSYx527Controller::updateStatusConnection, this);

	registerProgression(m_Driver, &CAENSYx527Controller::updateProgression, this);


}


void CAENSYx527Controller::refreshNTotalChannelProperty(int32 value) throw (CannotSetValue){
	islot.resize(value);
	ichannel.resize(value);
	namechannel.resize(value);
	i0set.resize(value);
	v0set.resize(value);
	imon.resize(value);
	vmon.resize(value);
	power.resize(value);
	statuschannel.resize(value);
	rampup.resize(value);
	rampdown.resize(value);
	isselected.resize(value);
}

void CAENSYx527Controller::refreshChannelNameProperty(int32 index, const string& value) throw (CannotSetValue) {

	if (value.length() > MAX_CH_NAME) {
		cout << "Channel name is too long, it must be < " << MAX_CH_NAME << endl;
	}
	else{
	// Set drivers channel name
		int flag=0;
		for(int i=0;i<ntotchannel();i++){
			if(i!=index && value==namechannel(i)){
				flag=1;
				cout << "Name already used for channel " << i << endl;
				break;
			}
		}
		if(value!=namechannel(index)&&!flag){
			m_Driver->writeParam(index,"Name",0.);
			log(Level::s_Info) << "High voltage " << namechannel(index) << " (slot=" << islot(index) <<",channel=" << ichannel(index)<<") renamed with " << value << endlog;
		}
	}
}

void CAENSYx527Controller::refreshV0SetProperty(int32 index,float64 value) throw (CannotSetValue) {
	if (value > vmax()) {
			cout << "V0Set " << value << " too high, must be < " << vmax() << " stay with "<< m_Driver->v0set(index) << endl;
	}
	else if (value < 0) {
			cout << "V0Set must be positive" <<" stay with "<< m_Driver->v0set(index) << endl;

	}
	else{
		  //cout << "refreshV0SetProperty : v0set=" << v0set(index) << " value=" << value << endl;

		  if(value!=v0set(index)){
			//  if(!m_Driver->locked){
				  m_Driver->writeParam(index,"V0Set",(float)value);
				  log(Level::s_Info) << "High voltage " << namechannel(index) << " (slot=" << islot(index) <<",channel=" << ichannel(index)<<") set to V0set=" << value << endlog;
			//  }
			//  else log(Level::s_Info) << "V0set cannot be changed during acquisition" << endlog;
		  }
	}
}

void CAENSYx527Controller::refreshI0SetProperty(int32 index,float64 value) throw (CannotSetValue) {
	if (value > 1000) {
			cout << "I0Set " << value << " too high, must be < 1000" << endl;
	}
	else if (value < 0) {
			cout << "I0Set must be positive" << endl;
	}
	else{
		    if(value!=i0set(index)){
		    	log(Level::s_Info) << "High voltage " << namechannel(index) << " (slot=" << islot(index) <<",channel=" << ichannel(index)<<") set to I0set=" << value << endlog;
		    	m_Driver->writeParam(index,"I0Set",(float)value);
		    }
	}
}

void CAENSYx527Controller::refreshRampUpProperty(int32 index,float64 value) throw (CannotSetValue) {
	if (value > 1000) {
			cout << "RampUp " << value << " too high, must be < 1000" << endl;
	}
	else if (value < 0) {
			cout << "RampUp must be positive" << endl;
	}
	else{
		if(value!=rampup(index)){
			m_Driver->writeParam(index,"RUp",(float)value);
		}
	}
}

void CAENSYx527Controller::refreshRampDownProperty(int32 index,float64 value) throw (CannotSetValue) {
	if (value > 1000) {
				cout << "RampDown " << value << " too high, must be < 1000" << endl;
		}
		else if (value < 0) {
				cout << "RampDown must be positive" << endl;
		}
		else{
			if(value!=rampdown(index)){
			    m_Driver->writeParam(index,"RDWn",(float)value);
			}
		}
}

void CAENSYx527Controller::refreshPowerProperty(int32 index,const string& value) throw (CannotSetValue) {

	if(value=="On"){
		if(!m_Driver->locked){
			m_Driver->powerOn(index);
			log(Level::s_Info) << "High voltage " << namechannel(index) << " (slot=" << islot(index) <<",channel=" << ichannel(index)<<") switched on" << endlog;
		}
		else log(Level::s_Info) << "HV channel cannot be switched on during acquisition" << endlog;
	}
	else{
		if(!m_Driver->locked){
			m_Driver->powerOff(index);
			log(Level::s_Info) << "High voltage " << namechannel(index) << " (slot=" << islot(index) <<",channel=" << ichannel(index)<<") switched off" << endlog;
		}
		else log(Level::s_Info) << "HV channel cannot be switched off during acquisition" << endlog;
	}

}
void CAENSYx527Controller::refreshVMaxProperty(float64 value) throw (CannotSetValue){

}

void CAENSYx527Controller::refreshDeltaVProperty(float64 value) throw (CannotSetValue){
	if(value!=0){
		for(int i=0;i<ntotchannel();i++){
		   v0set.set(i,v0set(i)+value);
		}

	}
}

/*
 * updateActualCurrent
 */
void CAENSYx527Controller::updateIMon(int32 index) {
	imon.set(index,m_Driver->imon(index));
}

/*
 * updateActualVoltage
 */
void CAENSYx527Controller::updateVMon(int32 index) {
	vmon.set(index,m_Driver->vmon(index));
}

/*
 * updateChannelName
 */
void CAENSYx527Controller::updateChannelName(int32 index) {
	namechannel.set(index,m_Driver->namechannel.get(index).c_str());
	//cout << "test update Channel Name " << index << endl;
}

/*
 * updateV0Set
 */
void CAENSYx527Controller::updateV0Set(int32 index) {
	v0set.update(index,m_Driver->v0set(index));
}

/*
 * updateI0Set
*/
void CAENSYx527Controller::updateI0Set(int32 index) {
	i0set.update(index,m_Driver->i0set(index));
}
/*
 * updatePower
 */
void CAENSYx527Controller::updatePower(int32 index) {
	power.update(index,m_Driver->power(index));
}
/*
 * updateRampUp
 */
void CAENSYx527Controller::updateRampUp(int32 index) {
	rampup.update(index,m_Driver->rampup(index));
}
/*
 * updateRampDown
 */
void CAENSYx527Controller::updateRampDown(int32 index) {
	rampdown.update(index,m_Driver->rampdown(index));
}
/*
 * updateStatus
 */
void CAENSYx527Controller::updateChannelStatus(int32 index) {
   statuschannel.set(index,m_Driver->statuschannel(index));
   if(m_Driver->statuschannel(index)==STATUSCHANNEL[3]||m_Driver->statuschannel(index)==STATUSCHANNEL[4]
	 ||m_Driver->statuschannel(index)==STATUSCHANNEL[5]||m_Driver->statuschannel(index)==STATUSCHANNEL[6]
	 ||m_Driver->statuschannel(index)==STATUSCHANNEL[7]||m_Driver->statuschannel(index)==STATUSCHANNEL[8]
	 ||m_Driver->statuschannel(index)==STATUSCHANNEL[9]){
	   	   statusall=false;
   }
   else if(statusall()==false){
	   int i=0;
	   while(i<ntotchannel()&&(m_Driver->statuschannel(index)!=STATUSCHANNEL[3]||m_Driver->statuschannel(index)!=STATUSCHANNEL[4]
			||m_Driver->statuschannel(index)!=STATUSCHANNEL[5]||m_Driver->statuschannel(index)!=STATUSCHANNEL[6]
			||m_Driver->statuschannel(index)!=STATUSCHANNEL[7]||m_Driver->statuschannel(index)!=STATUSCHANNEL[8]
			||m_Driver->statuschannel(index)!=STATUSCHANNEL[9])){
		   i++;
	   }
	   if(i==ntotchannel()) statusall=true;
   }

}
/*
 * updateStatus
 */
void CAENSYx527Controller::updateStatusConnection() {
	statusconnection.set(m_Driver->statusconnection());
}
/*
 * updateProgression
 */
void CAENSYx527Controller::updateProgression() {
	// Calculate its own progression to manage retry case
	int32 progression = m_Driver->commandProgression();
	if (progression >= 100) {
		progression = 99;
	}
	commandProgression = progression;
}

/*
 * read
 */
void CAENSYx527Controller::read() {

}

/*
 * stop
 */
void CAENSYx527Controller::stop() {

}

/*
 * start
 */
void CAENSYx527Controller::start() {


}

/*
 * on
 */
void CAENSYx527Controller::on() {
	for(int i=0;i<ntotchannel();i++){
		if(isselected(i)==true){
			if(!m_Driver->locked){
				m_Driver->powerOn(i);
				log(Level::s_Info) << "High voltage " << namechannel(i) << " (slot=" << islot(i) <<",channel=" << ichannel(i)<<") switched on" << endlog;
			}
			else log(Level::s_Info) << "HV channel cannot be switched on during acquisition" << endlog;
		}
	}
}

/*
 * off
 */
void CAENSYx527Controller::off() {
	for(int i=0;i<ntotchannel();i++){
		if(isselected(i)==true){
			if(!m_Driver->locked){
				m_Driver->powerOff(i);
				log(Level::s_Info) << "High voltage " << namechannel(i) << " (slot=" << islot(i) <<",channel=" << ichannel(i)<<") switched on" << endlog;
			}
			else log(Level::s_Info) << "HV channel cannot be switched off during acquisition" << endlog;
		}
	}
}
/*
 * select channels
 */
void CAENSYx527Controller::select() {


	if(selectchoice()=="all"){
		for(int i=0;i<ntotchannel();i++){
			isselected.set(i,true);
		}
	}
	else if(selectchoice()=="none"){
		for(int i=0;i<ntotchannel();i++){
			isselected.set(i,false);
		}
	}
	else if(selectchoice()=="slot"){
		for(int i=0;i<ntotchannel();i++){
			if(islot(i)==selectslot()) isselected.set(i,true);
			else isselected.set(i,false);
		}
	}
	else if(selectchoice()=="name"){
		for(int i=0;i<ntotchannel();i++){
			if(namechannel(i).find(selectstring())!=std::string::npos) isselected.set(i,true);
			//else isselected.set(i,false);
		}
	}


}

void CAENSYx527Controller::set() {
	for(int i=0;i<ntotchannel();i++){
		if(isselected(i)==true){
			//cout << setparam() << " " << setvalue() << endl;
			if(setparam()=="V0Set") v0set.set(i,strtof(setvalue().c_str(),NULL));
			else if(setparam()=="I0Set") i0set.set(i,strtof(setvalue().c_str(),NULL));
			else if(setparam()=="RUp") rampup.set(i,strtof(setvalue().c_str(),NULL));
			else if(setparam()=="RDWn") rampdown.set(i,strtof(setvalue().c_str(),NULL));
			else if(setparam()=="DeltaV") v0set.set(i,v0set(i)+strtof(setvalue().c_str(),NULL));

		}
	}
}

}

